# -*- coding: utf-8 -*-
""" CalvOS Logging System.

Module for the logging system use across all calvos engine.

@author: Carlos Calvillo
@copyright:  2020 Carlos Calvillo. All rights reserved.
@license:    GPL v3
"""
__version__ = '0.0.1'
__date__ = '2020-12-06'
__updated__ = '2020-12-06'

import logging
import traceback
from lxml import etree as ET
import xml.dom.minidom

# #TODO: Solve how to pass the log file path dynamically to this module

# A new handler to store "raw" LogRecords instances
class LogRecordsHandler(logging.Handler):
    """
    A handler class which stores LogRecord entries in a list
    """
    def __init__(self, records):
        """
        Initiate the handler
        :param records_list: a list to store the LogRecords entries
        """
        self.records = records
        super().__init__()

    def emit(self, record):
        self.records.append(record)
        

class Logger:
    """ A class for modeling a logger.
    """
    def __init__(self, name, level = logging.DEBUG, \
             abstract_list = None, to_console = True, \
             file_handler = None, formatter_str = None):
        
        self.name = name
        self.level = level
        self.to_console = to_console
        self.abstract_list = abstract_list
        
        #Create logger
        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)

        if formatter_str is not None:
            self.formatter = logging.Formatter(formatter_str)
        else:
            self.formatter = None
        
        #Add Handlers
        if to_console is True:
            pass
#            self.console_handler = logging.StreamHandler(sys.stdout)
#            if self.formatter is not None:
#                self.console_handler.setFormatter(formatter)
#            self.logger.addHandler(self.console_handler)
        if type(abstract_list) is list:
            self.logger.addHandler(LogRecordsHandler(abstract_list))
        if file_handler is not None:
            self.file_handler = file_handler
            if self.formatter is not None:
                self.file_handler.setFormatter(self.formatter)
            self.logger.addHandler(self.file_handler)
        

class Log:
    
    def __init__(self, level = logging.DEBUG, \
          log_file = None, formatter_str = None, \
          to_console = True, to_abstract = False):
    
        """ Class to model a log system for a calvOS project """
        self.level = level
        self.to_console = to_console
        
        self.loggers = {}   #{name, Logger object}
        
        self.abstract_logs = []
        
        self.default_formatter_str = \
            '[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s'
        
        if formatter_str is None:
            self.formatter_str = self.default_formatter_str
        else:
            self.formatter_str = formatter_str
        self.formatter = logging.Formatter(formatter_str)
        
        self.file_handler = None 
        if log_file is not None:
            self.file_handler = logging.FileHandler(log_file)
            self.file_handler.setLevel(level)
            self.file_handler.setFormatter(self.formatter)
        
        self.counters = {"debug" : 0,
                         "info" : 0,
                         "warning" : 0,
                         "error" : 0,
                         "critical" : 0}
            
    def set_output_file(self, file_name):
        self.file_handler = logging.FileHandler(file_name)
    
    def set_formatter(self,formatter_str = None):
        if formatter_str is not None:
            self.formatter = logging.Formatter(formatter_str)
        else:
            self.formatter = self.default_formatter_str
            
    
    def set_abstract_logs(self, abstract_list):
        if type(abstract_list) is list:
            self.abstract_logs = abstract_list

    def add_logger(self, name, level = None, \
                   abstract_logs = None, \
                   to_console = None, file_handler = None, formatter_str = None):
        """ Adds a new logger to the logging system. """

        if level is None:
            level = self.level
        if abstract_logs is None:
            abstract_logs = self.abstract_logs
        if to_console is None:
            to_console = self.to_console
        if file_handler is None:
            file_handler = self.file_handler
        if formatter_str is None:
            formatter_str = self.formatter_str
                   
        self.loggers.update( {name : Logger(name, level, \
             abstract_logs, to_console, file_handler, formatter_str)} )
    
    def debug(self, name, message):
        """ """
        self.loggers[name].logger.debug(message)
        self.counters["debug"] += 1
    
    def info(self, name, message):
        """ """
        self.loggers[name].logger.info(message)
        self.counters["info"] += 1
    
    def warning(self, name, message):
        """ """
        self.loggers[name].logger.warning(message)
        self.counters["warning"] += 1
    
    def error(self, name, message):
        """ """
        self.counters["error"] += 1
        if traceback.format_exc().find("NoneType") == -1:
            trace = traceback.format_exc()
            print("ERROR: ", message)
            traceback.print_exc()
        else:
            trace = traceback.format_stack()
            print("ERROR: ", message)
            traceback.print_stack()
        message += "\nError traceback/callstack:\n"
        if type(trace) is list:
            for entry in trace:
                message += entry
        else:
            message += str(trace)
            
        self.loggers[name].logger.error(message)

    def critical(self, name, message):
        """ """
        self.loggers[name].logger.critical(message)
        self.counters["critical"] += 1
    
    def to_xml(self, output_file = None):
        #Generate Logs root node and its information
        XML_root = ET.Element("Logs")
        for record in self.abstract_logs:
            XML_level_1 = ET.Element("Logger")
            XML_level_1.text = record.name
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("Time")
            XML_level_1.text = record.asctime
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("Filename")
            XML_level_1.text = record.filename
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("lineNo")
            XML_level_1.text = record.lineno
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("functionName")
            XML_level_1.text = record.funcName
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("LevelName")
            XML_level_1.text = record.levelname
            XML_root.append(XML_level_1)
            XML_level_1 = ET.Element("Message")
            XML_level_1.text = record.getMessage()
            XML_root.append(XML_level_1)

        print("Generating XML...")
        XML_string = ET.tostring(XML_root)
        
        XML_string = xml.dom.minidom.parseString(XML_string)
        XML_string = XML_string.toprettyxml()

        if output_file is None:
            print(XML_string)
        else:
            myfile = open(output_file, "w")
            myfile.write(XML_string)


log_system = None




# -*- coding: utf-8 -*-
""" Time related utilities.

calvos.utils.time models utilities for time related function.

@author:     Carlos Calvillo
@copyright:  2021 Carlos Calvillo. All rights reserved.
@license:    GPL v3
"""
__version__ = '0.0.1'
__date__ = '2021-01-17'
__updated__ = '2021-01-17'
    
import pyexcel as pe
import math
from lxml import etree as ET
import xml.dom.minidom
import pickle as pic
import cogapp as cog
import pathlib as pl
import importlib
import json

import calvos.common.codegen as cg
import calvos.common.logsys as lg
import calvos.common.general as grl

# --------------------------------------------------------------------------------------------------
# Definitions for the logging system
# --------------------------------------------------------------------------------------------------
LOGGER_LABEL = "time"

log = lg.log_system
log.add_logger(LOGGER_LABEL)

def log_debug(message):
    """ Wrapper for logging a debug message. """
    
    log.debug(LOGGER_LABEL, message)
    
def log_info(message):
    """ Wrapper for logging an info message. """
    
    log.info(LOGGER_LABEL, message)
    
def log_warn(message):
    """ Wrapper for logging a warning message. """
    
    log.warning(LOGGER_LABEL, message)
    
def log_error(message):
    """ Wrapper for logging an error message. """
    
    log.error(LOGGER_LABEL, message)
    
def log_critical(message):
    """ Wrapper for logging a critical error message. """
    
    log.critical(LOGGER_LABEL, message)
# --------------------------------------------------------------------------------------------------
# Cog sources to use for code generation of this module
cog_sources = cg.CogSources("utils.time")

class Timers():
    """ Models Utilities for the calvos projects. """
    def __init__(self, project_obj, **kwargs):
        
        self.project_obj = project_obj
        self.name = kwargs.get('name', None)
        self.input_file = kwargs.get('input_file', None)
        
        self.timers = {} # Expects { timer_name : Timer object}
        self.tasks = []
        
        self.module = "utils.time"
        
        self.gen_path = self.project_obj.get_component_gen_path(self.module)
        
        self.metadata_gen_source_file = None
        
    def update_timer_tasks(self):
        """ Define which timer tasks to use for timers handling. """
        self.tasks = []
        if self.base_tick_ms is not None:
            for task_option in self.tasks_options:
                # Task option shall be equal to or greater than the base tick.
                if task_option >= self.base_tick_ms:
                    # Get an approximation to the option value
                    aprox_value = int(round(task_option / self.base_tick_ms) * self.base_tick_ms)
                    self.tasks.append(aprox_value)
                    if task_option % self.base_tick_ms != 0:
                        log_warn(("Timer task cycle time was approximated." \
                                  + "Target: %fms, actual: %fms, base tick: %fms.") \
                                  % (task_option, aprox_value, self.base_tick_ms))
        else:
            log_warn("Can't define timer tasks since there is no base tick defined.")
            
    #===============================================================================================    
    def parse_spreadsheet_ods(self,input_file):
        """ parses an spreadsheet (ods format) to generate the timers. """
        book = pe.get_book(file_name=str(input_file))
        self.metadata_gen_source_file = str(input_file)
        
        self.input_file = cg.string_to_path(input_file)
        
        
        # -----------------------
        # Parse Parameters
        # -----------------------
        log_debug("Parsing parameters data...")
        CONFIG_TITLE_ROW = 1 # Row number with titles
        working_sheet = book["Config"]
        working_sheet.name_columns_by_row(CONFIG_TITLE_ROW)
        
        for idx, row in enumerate(working_sheet):
            # Ignore rows that are before the "config's" title row
            if idx >= CONFIG_TITLE_ROW:
                param_id = row[working_sheet.colnames.index("Parameter")]
                param_value = row[working_sheet.colnames.index("User Value")]
                if param_id != "" \
                and self.project_obj.simple_param_exists(self.module, param_id) is True:
                    read_only = self.project_obj.simple_param_is_read_only(self.module, param_id)
                    if read_only is False and param_value != "":
                        # Add parameter user's value to this object
                        param_type = self.project_obj.get_simple_param_type(self.module, param_id)
                        # Force read value from ODS to be a string and substitute quotes characters
                        # out of utf-8 if present.
                        param_value = str(param_value)
                        param_value = param_value.encode(encoding='utf-8')
                        param_value = param_value.replace(b'\xe2\x80\x9c', b'\"')
                        param_value = param_value.replace(b'\xe2\x80\x9d', b'\"')
                        param_value = param_value.decode('utf-8')
                        log_debug("Processing parameter '%s' with value '%s'..." \
                                  % (param_id, param_value))
                        param_value = grl.process_simple_param(param_type, param_value)
                        
                        # Check if validation is required
                        validator = \
                            self.project_obj.get_simple_param_validator(self.module, param_id)
                        if validator is not None:
                            # Validate parameter
                            valid = self.project_obj.validate_simple_param(self.module, param_id, \
                                                                           validator[0], \
                                                                           validator[1])
                        else:
                            # If no validator is defined assume data is valid.
                            valid = True
                        
                        if valid is True:
                            param_obj = grl.SimpleParam(param_id, param_type, param_value)
                            self.simple_params.update({param_id: param_obj})
                            log_debug("Added CAN user parameter '%s', type '%s' with value '%s'" \
                                      % (param_id, param_type, param_value))
                        else:
                            log_warn(("Parameter '%s:%s' is invalid as per its " \
                                     + "validator '%s:%s''. Parameter ignored.") \
                                     % (self.module, param_id, validator[0], validator[1]))
                    elif param_value != "":
                        log_warn("Parameter '%s' is read-only. Ignored user value." % param_id)
                    else:
                        pass
                elif param_id != "":
                    log_warn("Parameter '%s' is meaningless for component '%s'. Parameter ignored." \
                             % (param_id, self.module))
                else:
                    pass
        
        log_debug("Loading Timers information from file: '%s'..." % input_file)
        # -----------------------
        # Parse Timers Data
        # -----------------------
        working_sheet = book["Timers"]
        TIMERS_TITLES_ROW = 0 # Row where the node's titles are located
        working_sheet.name_columns_by_row(TIMERS_TITLES_ROW)
        
        for idx, row in enumerate(working_sheet):
            # Ignore rows that are before the "node's" title row
            if idx >= TIMERS_TITLES_ROW:
                timer_obj = None
                timer_def_error = False
                timer_id = row[working_sheet.colnames.index("Timer Id")]
                # Skip rows with no timer id
                
                if str(timer_id) != "": 
                    log_debug("Adding Timer '%s'..." % str(timer_id))
                    if cg.is_valid_identifier(timer_id) is True:
                        timer_obj = self.Timer(timer_id)
                    else:
                        log_warn(("Timer id '%s' is not valid. It shall comply with C identifier " \
                                  + "syntax. Timer ignored.") % timer_id)
                
                if timer_obj is not None:
                    # Gather rest of timer parameters
                    timer_obj.desc = str(row[working_sheet.colnames.index("Description")])
                    # Get timer tick base time (ms)
                    try:
                        base_time = int(row[working_sheet.colnames.index("Base tick time (ms)")])
                        # Confirm that selected base tick is a valid value 
                        validator = self.project_obj.get_simple_param_val("utils.time","time_tasks")
                        if base_time in validator:
                            timer_obj.base_time = base_time
                        else:
                            timer_def_error = True
                            log_warn(("Invalid timer '%s' base time. Allowed values are: '%s'. " \
                                      + "Timer not added.") % (timer_id,validator))
                    except Exception as e:
                        timer_def_error = True
                        log_warn(("Timer '%s' base time shall be an integer. Timer not added." \
                                 + ". Raised exception: \n%s") % (timer_id,e))
                    if timer_def_error is False:
                        # Get timer type if specified.
                        timer_type = str(row[working_sheet.colnames.index("Type")])
                        validator = self.project_obj.get_simple_param_val("utils.time","time_types")
                        default = self.project_obj.get_simple_param_val("utils.time", \
                                                                        "time_types_default")
                        if timer_type == "":
                            log_warn("Assumed type for timer '%s': '%s'." \
                                     % (timer_id, default))
                        else:
                            if timer_type in validator:
                                timer_obj.type = timer_type
                            else:
                                timer_def_error = True
                                log_warn(("Invalid timer '%s' type. Allowed values are: '%s'. " \
                                          + "Timer not added.") % (timer_id,validator))
                    if timer_def_error is False:
                        auto_reload = str(row[working_sheet.colnames.index("Auto reload")])
                        if auto_reload == "True" or auto_reload == "true" or auto_reload == "TRUE":
                            timer_obj.auto_reload = True
                        elif auto_reload == "False" or auto_reload == "false" \
                        or auto_reload == "FALSE":
                            timer_obj.auto_reload = False
                        else:
                            timer_obj.auto_reload = False
                            log_warn("Timer '%s' Auto reload shall be boolean. Assumed 'False'." \
                                     % timer_id)
                    if timer_def_error is False:
                        # Get timer type if specified, otherwise assume automatic.
                        reloads = str(row[working_sheet.colnames.index("Reloads")])
                        if reloads == "":
                            if timer_obj.auto_reload is False:
                                # If auto reload is False we need a defined number of reloads
                                # including zero.
                                timer_obj.reloads = reloads = 0
                                log_warn("Assumed zero reloads for timer '%s'" % timer_id)
                        else:
                            try:
                                reloads = \
                                    int(row[working_sheet.colnames.index("Reloads")])
                                timer_obj.reloads = reloads 
                            except Exception as e:
                                timer_def_error = True
                                log_warn(("Timer '%s' reloads shall be an integer. Timer not " \
                                         + "added. Raised exception: \n%s") % (timer_id,e))
                    if timer_def_error is False:
                        period = str(row[working_sheet.colnames.index("Period (ms)")])
                        if period != "" and timer_obj.auto_reload is True \
                        or timer_obj.reloads > 0:
                            period_mandatory = True
                        else:
                            period_mandatory = False
                        try:
                            period = int(row[working_sheet.colnames.index("Period (ms)")])
                            timer_obj.period = period
                        except Exception as e:
                            if period_mandatory:
                                log_warn(("Timer '%s' period shall be an integer if defined as " \
                                          + "auto-reload or has reloads > 0. Timer not " \
                                          + "added. Raised exception: \n%s") % (timer_id,e))
                            else:
                                log_info(("Timer '%s' period shall be an integer. " \
                                          + "Timer may be added with no period.") % timer_id)
                    if timer_def_error is False:
                        auto_start = str(row[working_sheet.colnames.index("Auto start")])
                        if auto_start == "True" or auto_start == "true" or auto_start == "TRUE":
                            timer_obj.auto_start = True
                        elif auto_start == "False" or auto_start == "false" \
                        or auto_start == "FALSE":
                            timer_obj.auto_start = False
                        else:
                            timer_obj.auto_start = False
                            log_warn("Timer '%s' Auto start shall be boolean. Assumed 'False'." \
                                     % timer_id)
                    
                    # If no critical error found and timer is not duplicated, 
                    # then add Timer to the Timers object.
                    if timer_id not in self.timers:
                        self.timers.update({timer_id : timer_obj})
                    else:
                        log_warn("Timer '%s' is duplicated. Duplications not added." % timer_id)
            
        
    class Timer():
        """ Models a timer. """
        def __init__(self, id_str, **kwargs):
            self.id_str = id_str
            
            self.desc = kwargs.get('desc', None)
            
            self.time_ticks = kwargs.get('time_ticks', None)
            if self.time_ticks is None:
                self.time_ms = kwargs.get('time_ms', None)
                if self.time_ms is not None:
                    self.set_time_ticks_from_ms(self.time_ms)
                    
            callback_str = kwargs.get('callback', None)
            if callback_str is not None and cg.is_valid_identifier(callback_str):
                self.callback = callback_str
            
            # callback_cycle triggers when a cycle finishes, for example if timer auto-reloads
            # 10 times, it will be called once the 10 times elapsed.
            # TODO: This can be a flag instead of a callback.
            callback_str = kwargs.get('callback_cycle', None)
            if callback_str is not None and cg.is_valid_identifier(callback_str):
                self.callback_cycle = callback_str
                
            self.reloads = kwargs.get('reloads', 0) # 0...n or "inf"
            self.auto_reload = kwargs.get('auto_reload', False)
            
            self.type = kwargs.get('type', None)
            
            self.base_time = kwargs.get('base_time', None) 
            
            self.auto_start = kwargs.get('auto_start', None)
        
        def get_task_tick_time(self, timer_time_ms):
            """ Returns the cycle time of the closes task for processing the given timer value. """
            pass
                    
        def set_time_ticks_from_ms(self, time_ms):
            """ Sets the time in ticks from an input in milliseconds. """
             
            if self.base_tick_ms is not None:
                if time_ms & self.base_tick_ms != 0:
                    self.time_ticks = int(round(time_ms / self.base_tick_ms))
                    log_warn(("Timer '%s' rounded. Requested time %fms, rounded time %dms." \
                              + " Current base tick value: %f. ") \
                              % (self.name, time_ms, self.base_tick_ms))
                else:
                    self.time_ticks = int(time_ms / self.base_tick_ms)
            else:
                log_warn(("Can't set value of timer '%s' because there is no base tick defined." \
                          + " Define first the value of the timers base tick.") % self.name)
        
        def determine_timer_type(self):
            pass
        
        #===============================================================================================        
    def update_cog_sources(self):
        """ Updates the list of cog file(s) for this module. """
        
        global cog_sources
        cog_sources = cg.CogSources(self.module, gen_path = self.gen_path)
        
        self.add_cog_source("time_h", "cog_utils_time.h", True)
        self.add_cog_source("time_core_h", "cog_utils_time_core.h", True, \
                            [[self.module, "time_h"]])
        self.add_cog_source("time_core", "cog_utils_time_core.c", False, \
                            [[self.module, "time_core_h"]])

    #===============================================================================================    
    def get_cog_includes(self, source_id): 
        """ Returns a list of strings to be used with #include 'string'. """
        
        global cog_sources 
        return_str = []
        if source_id in cog_sources.sources:
            for relation in cog_sources.sources[source_id].relations:
                target_module = importlib.import_module("calvos."+relation.module)
                target_srcs = target_module.cog_sources
                if relation.source_id in target_srcs.sources:
                    return_str.append(target_srcs.sources[relation.source_id].cog_out_file)
                else:
                    log_warn("Include file id '%s' of module '%s' not found." \
                             % (source_id, "calvos."+relation.module))
        else:
            log_warn("Include file id '%s' for CAN module not defined.")
        
        return return_str

    #===============================================================================================    
    def add_cog_source(self, cog_id, cog_in_file, is_header = False, relations = None, \
                       dparams = {}):
        """ Adds a cog source to this network object's cog files list. 
        
        Parameters
        ----------
            relations, list
                Should be a list of lists to model the relation of cog_id source file with other
                source files... expected -> [[module2, source_id1], [modul2, source_id2], ...]
        """ 
        
        global cog_sources

        ##### cog_out_file = self.get_cog_out_name(cog_in_file, self.id_string)
        cog_out_file = self.get_cog_out_name(cog_in_file)
        cog_sources.sources.update({cog_id: cg.CogSources.CogSrc(\
                cog_id,cog_in_file, cog_out_file, is_header, [], dparams)})
        
        if relations is not None:
            for relation in relations:
                relation_obj = cg.CogSources.CogSrcRel(relation[0],relation[1])
                cog_sources.sources[cog_id].relations.append(relation_obj)
    
    #===============================================================================================    
    def get_cog_out_name(self, cog_in_file, component_name = None): 
        """ Generates cog output file name as per a given input file name. """
        #remove "cog_" prefix to output file names
        if cog_in_file.find("cog_") == 0:  
            cog_output_file = cog_in_file[4:]  
        #substitute component name if found (COMP)
        if component_name is not None:
            cog_output_file = cog_output_file.replace('COMP',component_name)
        else:
            cog_output_file = cog_output_file.replace('COMP','')
        
        return cog_output_file
        
    #===============================================================================================     
    def cog_generator(self, input_file, cog_output_file, project_pickle_file, variables = None):
        """ Invoke cog generator for the specified file.
        """
        # Setup input and output files
        # ----------------------------
        out_dir = self.project_obj.get_simple_param_val("common.project", "project_path_out")
        work_dir = self.project_obj.get_simple_param_val("common.project", "project_path_working")
        gen_path = self.gen_path
        
        comgen_CAN_cog_input_file = gen_path / str(input_file)
                
        # Set output file with path
        comgen_CAN_cog_output_file = out_dir / cog_output_file
                
        # Invoke code generation
        # ----------------------           
        cog_arguments = ['dummy_argument', \
                   '-d', \
                   '-D', 'input_worksheet=' + self.metadata_gen_source_file, \
                   '-D', 'project_working_dir=' + str(work_dir), \
                   '-D', 'cog_proj_pickle_file=' + str(project_pickle_file), \
                   '-D', 'cog_output_file=' + str(cog_output_file)]
        
        # Append additional variables if required
        if variables is not None:
            for variable, variable_val in variables.items():
                variable_str = str(variable) + "=" +str(variable_val)
                cog_arguments.append('-D')
                cog_arguments.append(variable_str)

        cog_arguments.append('-o')
        cog_arguments.append(str(comgen_CAN_cog_output_file))
        cog_arguments.append(str(comgen_CAN_cog_input_file))
        
                
        # Call cogapp engine
        cog_return = cog.Cog().main(cog_arguments)
        if cog_return == 0:
            log_info("Code generation successful: '%s'" % comgen_CAN_cog_output_file)
            print("INFO: code generation successful: ",comgen_CAN_cog_output_file)
        else:
            log_error("Code generation error for '%s'. Cogapp return code: '%s'" \
                 % (comgen_CAN_cog_output_file, str(cog_return)))
            print("INFO: code generation return value: ",cog_return)
        
    #===============================================================================================   
    def gen_code(self):
        """ Generates C code for the given network.
        """

        global cog_sources
                
        # Get project pickle full file name 
        project_pickle = self.project_obj.get_work_file_path("common.project","project_pickle")
         
        #----------------------------------------------------------------------
        # Generate source file(s)
        #----------------------------------------------------------------------
        for cog_source in cog_sources.sources.values():
            # Generate includes variable if needed
            variables = None
            includes_lst = self.get_cog_includes(cog_source.source_id)
            if len(includes_lst) > 0:
                variables = {}
                include_var = json.dumps(includes_lst)
                variables.update({"include_var" : include_var})
                  
            self.cog_generator(cog_source.cog_in_file, cog_source.cog_out_file, project_pickle, \
                              variables)
#         
#         # Delete pickle file
#         # ------------------
#         print("INFO: deleting pickle file")
#         cg.delete_file(cog_serialized_network_file)
#         
#         #TODO: Logic for ensuring the current network is meaningful, e.g., is no emtpy, etc.
#         if node_names == None:
#             # C code for all nodes will be generated.
#             pass
#         elif len(node_names) > 0:
#             # Generate C code for specified node each
#             for node in node_names:
#                 #Check if node exists
#                 if node in self.nodes:
#                     pass
#                 else:
#                     pass
#         else:
#             log_warn(("Invalid input. No C-Code will be generated for " 
#                           + "network: " + self.name + "."))

#===================================================================================================
def load_input(input_file, input_type, params, project_obj):
    """ Loads input file and returns the corresponding object.
    
    This function is invoked from the project handler.
    
    Parameters
    ----------
        input_file : Path
            Path object pointing to the input file for the current project' module.
        input_type : int
            Type of input, can be either IN_TYPE_XML or IN_TYPE_ODS. Those constants are
            defined in the "project" module.
        params : dict
            Dictionary of parameters if required.
    
    Returns
    -------
        Object
        This function returns an object with the loaded data for the current module. For CAN
        Networks this object is of class Network_CAN.
    """
    del input_type, params # Unused parameters
    try:
        return_object = Timers(project_obj)
        return_object.parse_spreadsheet_ods(input_file)
        try:
            return_object.update_cog_sources()
        except Exception as e:
            log_error('Failed to update cog sources. Reason: %s' % (e))
    except Exception as e:
        return_object = None
        log_error('Failed to process input file "%s". Reason: %s' % (input_file, e))
    finally:
        return return_object    
    
    return return_object

#===================================================================================================
def generate(input_object, out_path, working_path, calvos_path, params = {}):
    """ Generate C code for the given object.
    
    Parameters
    ----------
        input_object : Object
            Input object to use for the code generation of this module.
        out_path : Path
            Output path were the generated code is to be placed.
        working_path : Path
            Working path to be used for storage of temporary data used during code
            generation.
        calvos_path : Path
            Path of the calvos python package
        params : dict
            Dictionary of parameters (if required)
    
    Yields
    ------
        This function produced C-code for the corresponding module. This code
        is generated in the provided out_path.
    """
    del params # Unused parameter
#     #TODO: Check for required parameters
# #     try:
#     cog_files_path = calvos_path / "comgen" / "gen" / "CAN"
    input_object.gen_code()
#     # Generate XML
#     xml_output_file = out_path / (str(input_object.input_file.stem) + ".xml")
#     input_object.gen_XML(xml_output_file)


# -*- coding: utf-8 -*-
""" CalvOS Project Manager.

Module for managing a calvos project.

@author: Carlos Calvillo
@copyright:  2020 Carlos Calvillo. All rights reserved.
@license:    GPL v3
"""
__version__ = '0.0.1'
from odf import element
__date__ = '2020-11-12'
__updated__ = '2020-11-12'

import xml.etree.ElementTree as ET
import xml.dom.minidom
import shutil
import pathlib
import importlib
import json
import re
import pickle

import calvos.common.codegen as cg
import calvos.common.logsys as lg
import calvos.common.general as grl


IN_TYPE_XML = 0
IN_TYPE_ODS = 1

# -----------------------------------------------------------------------------
# Definitions for the logging system
# -----------------------------------------------------------------------------
LOGGER_LABEL = "proj"

log = lg.log_system
log.add_logger(LOGGER_LABEL)

def log_debug(message):
    log.debug(LOGGER_LABEL, message)
    
def log_info(message):
    log.info(LOGGER_LABEL, message)
    
def log_warn(message):
    log.warning(LOGGER_LABEL, message)
    
def log_error(message):
    log.error(LOGGER_LABEL, message)
    
def log_critical(message):
    log.critical(LOGGER_LABEL, message)
# -----------------------------------------------------------------------------

#===================================================================================================
class Project:
    """ Class to mode a calvos project. """

    #===============================================================================================
    def __init__(self, name, project_file, calvos_path):
        """ Class constructor.
        
        Parameters
        ----------
            name : str
                Name of the project
            project_file : str
                Path and file name of the project to be processed.
            calvos_path : str
                Path of the calvos python package.
        """
        self.name = name
        self.calvos_path = cg.string_to_path(calvos_path)

        if project_file is not None:
            project_file = cg.string_to_path(project_file)
            self.project_file = project_file
            # Extract project path
            self.project_path = project_file.parent.resolve()
            
            self.paths = {"project_path" : cg.string_to_path(self.project_path),
                      "project_inputs" : self.project_path / "usr_in",
                      "project_output" : self.project_path / "out",
                      "project_working_dir" : self.project_path / "out/working_dir",
                      "project_docs" : self.project_path / "out/doc",
                      "project_templates" : self.project_path / "out/doc/templates",
                      "calvos_path" : cg.string_to_path(str(self.calvos_path))
                      }
        else:
            self.project_file = None
            self.project_path = None
            self.paths = {}
        
        self.components_definitions = {} #Expected dict {comp_type : CompDefinition object}
        self.components = [] #List of Component objects
        
        self.params = {} # param objects {param_id : param value}
        
        self.param_categories = {} # {category id : objects ParamCategory }
        
        self.simple_params = {}
        
        self.module = "common.project"
        
        
        
        self.test_dict = {"hola1" : 1, "hola2" : 2, "hola3" : 3, "hola4" : 4, "hola5" : 5}
        self.test_list = ["lst1", "lst2", "lst3", "lst4", "lst5"]
        self.test_var = 100
    
    #===============================================================================================
    def load_component_definitions(self, XML_root, comp_def_ojb):
        """ Loads the component definitions from the given xml file.
        comp_def_ojb: CompDefinition type """

        # Define XML namespace(s)
        nsmap = {"clv":"calvos"}
        
        component = comp_def_ojb.type
        
        log_debug("Loading definitions for component '%s'..." % component)
        
        # Load Title
        title = XML_root.find("./clv:Title", nsmap)
        if title is not None:
            comp_def_ojb.title = title.text
        
        # Load Description
        desc = XML_root.find("./clv:Desc", nsmap)
        if desc is not None:
            comp_def_ojb.desc = desc.text
                
        # Load required instances
        instances = XML_root.get("instances", None)
        if instances is not None:
            instances = grl.process_simple_param("list",instances)
            if instances is not None:
                comp_def_ojb.instances = instances
        
        # Load component definition inputs if any
        for input_file in XML_root.findall("./clv:Inputs/clv:Input", nsmap):
            data_input = str(input_file.text).strip('"')
            data_input = cg.string_to_path(data_input)
            if data_input is not None:
                input_type = input_file.get("type", None)
                if input_type is None:
                    log_warn(("Missing attribute 'type' for input data (element clv:Input) " \
                           + "in XML component definition of component '%s'.") \
                           % str(component))
                is_default = grl.process_simple_param("boolean", \
                                                      input_file.get("is_default", "False"))
                if is_default is True:
                    comp_def_ojb.default_input.update({data_input : input_type}) 
                else:
                    comp_def_ojb.inputs.update({data_input : input_type})
            else:
                log_error(("Wrong string '%s' for input data (element clv:Input) " \
                           + "in XML component definition of component '%s'.") \
                           % (str(input_file) , str(component)))
        
        # Load parameters if any
        for param in XML_root.findall("./clv:ParamsDefinitions/clv:ParamDefinition", nsmap):
            param_id = param.get("id", None)
            log_debug("Loading parameter '%s:%s'..." % (component, param_id))
            if param_id != None:
                param_type = param.get("type", None)
                if param_type is not None:
                    param_value = param.findtext("./clv:Default", None, nsmap)
                    if param_value is not None:
                        param_value = grl.process_simple_param(param_type, param_value)
                    else:
                        log_warn(("Missing default value for parameter '%s' of component '%s'." \
                                     + " Assumed 'None'.") % (param_id, component))
                    
                    param_read_only = param.get("is_read_only", None)
                    if param_read_only is None:
                        # If parameter is not found then it is not read-only
                        param_read_only = False
                    else:
                        param_read_only = grl.process_simple_param('boolean', param_read_only)
                        if param_read_only is None:
                            # If there is a problem converting read only parameter assume True
                            param_read_only = True
                            log_warn(("Wrong value for 'is_read_only' attribute of parameter '%s'" \
                                + " of component '%s'. Assumed read only.")%(param_id, component))
                    
                    param_validator = param.get("validator", None)
                    if param_validator is not None:
                        log_debug("Validating validator '%s'..." % param_validator)
                        param_validator = grl.process_simple_param('list', param_validator)
                        if param_validator is None:
                            # If there is a problem converting validator assume invalid
                            param_validator = None
                            log_warn(("Wrong value for 'validator' attribute of parameter '%s'" \
                                + " of component '%s'. Assumed 'None'.")%(param_id, component))

                    param_ojb = grl.SimpleParam(param_id, param_type, param_value, \
                                                    param_read_only, param_validator)
                    comp_def_ojb.simple_params.update({param_id : param_ojb})
                else:
                    log_warn(("Missing parameter type for parameter '%s' of component '%s'. " \
                              + "Component not added.") % (param_id, component))
            else:
                log_warn(("Id '%s' not found for a parameter of component '%s'. " \
                             + "Parameter not added.") % component)
        
        log_info("Loading of definitions for component '%s' completed." % component)   
   
    #===============================================================================================    
    def add_component(self, type_name, \
                     input_file = None, input_type = IN_TYPE_XML, params = {}, \
                     name = None, desc = None):
        """ Adds a Component to this project.
        
        Function load_component_definitions needs to be called prior to this one.
        """
        
        if type_name in self.components_definitions:
            component = self.Component(type_name, \
                         input_file, input_type, params, \
                         name, desc)
            
            self.components.append(component)
        else:
            log_warn("Undefined component type '%s'. Component not added." % type_name)
    
    #===============================================================================================        
    def find_components(self, _calvos_path):
        """ Search for calvos components within calvos package.
        
        Parameters
        ----------
            _calvos_path : Path
                Path of the calvos package.
        
        Returns
        -------
            If at least one component is found returns a dictionary with following format:
            {component_name1 : component_xml_path1, component_name2 : component_xml_path2, ...}
            If no component is found then returns an empty dictionary.
        """
        
        calvos_components = {} # {name : xml_path}
        for path in sorted(_calvos_path.rglob('*.py')):
            module_name = path.stem
            xml_path = path.with_suffix(".xml")
            # Check if corresponding xml file is found
            if xml_path.is_file() is True:
                # Python module is a calvos component, now gather its module string
                calvos_path_found = False
                module_name_list = []
                module_name_string = ""
                for parent in xml_path.parents:
                    if parent.name == "calvos":
                        calvos_path_found = True
                        break;
                    else:
                        module_name_list.append(str(parent.name))
                if calvos_path_found is True:
                    # Form module name string
                    for piece in module_name_list:
                        module_name_string = piece + "." + module_name_string
                    module_name_string += module_name
                    calvos_components.update({module_name_string : xml_path})
                    log_debug("Found component: '%s', XML: '%s'" \
                              % (module_name_string, calvos_components[module_name_string])) 
        
        return calvos_components
    
    #===============================================================================================  
    def load_project(self):
        """ Loads current project.
        
        Loading includes:
         - loading the parameter and component definitions (not project specific)
         - loading 
        """
        
        success = True
        
        # Set default project paths
        # -------------------------
        
        # Load project definitions
        # ------------------------
        #project_definitions_file = self.calvos_path / "common/project_definitions.xml"
        
        log_info('Loading project parameters and component definitions...')
        
        # Search for calvos components
        # ----------------------------
        # A calvos component shall have a module.py together with a module.xml at same path
        calvos_components = self.find_components(self.calvos_path)
        
        # Load parameters of found components
        # -----------------------------------
        for comp_name, comp_xml in calvos_components.items():
            # Create component definition object
            self.components_definitions.update({comp_name : self.CompDefinition(comp_name)})
            # Parse XML with component definitions
            XML_tree = ET.parse(comp_xml)
            XML_root = XML_tree.getroot()
            self.load_component_definitions(XML_root, self.components_definitions[comp_name]) 
        
        if success is True:
            log_info("============== Validating parameters' default data ==============")
            valid = self.validate_simple_params_defaults()
            if valid is False:
                log_warn("Invalid data found during parameters validation.")
                #TODO. determine wheter or not to continue processing, success = True/False? 
        
        if success is True:
            # Load project components
            log_info("============== Loading project components. ==============")
            
            project_file_path = self.project_file
            XML_tree = ET.parse(project_file_path)
            XML_root = XML_tree.getroot()
            
            for component in XML_root.findall("./Components/Component"):
                component_type = component.get("type")
                
                if component_type is not None \
                and component_type in self.components_definitions:
                
                    log_info("-------------- Loading component of type '%s'." \
                             % component_type)
                
                    component_name = str(component.findtext("Name"))
                    component_desc = str(component.findtext("Desc"))
                    component_input_type = str(component.find("Input").get("type"))
                    if component_input_type == "ods":
                        component_input_type = IN_TYPE_ODS
                    elif component_input_type == "xml":
                        component_input_type = IN_TYPE_XML
                    else:
                        component_input_type = None
                        log_warn('Invalid input type "%s" for component "%s"' \
                                  % (component_input_type, component_name))
                        
                    # Get input file, remove trailing/leading quotes
                    component_input = str(component.findtext("Input")).strip('"')
                    component_input = cg.string_to_path(component_input)
                    # Check if path is relative and make it aboslute based on project path
                    if component_input.is_absolute() is False:
                        component_input = self.paths["project_path"] / component_input
                        
                    component_params = {}
                    for param in component.findall("./Params/Param"):
                        param_name = param.get("name")
                        param_value = str(param.text)
                        if param_name is not None:
                            component_params.update( \
                                { param_name : param_value })
                        else:
                            log_warn('Invalid parameter name for component "%s".' % param_name)
                    
                    project_component = self.Component(component_type, component_input, \
                                                       component_input_type, component_params, \
                                                       component_name, component_desc)
                    
                    self.components.append(project_component)

                else:
                    log_warn('Component type is invalid: "%s".' % component_type)
            
            # Check if mandatory components are missing from project. In such case use default
            # input data.
            for component_def in self.components_definitions.values():
                if component_def.instances is not None \
                and len(component_def.instances) == 1 \
                and component_def.instances[0] == 1:
                # Component is mandatory. Check if user provided one
                    user_component_found = False
                    for component in self.components:
                        if component.type == component_def.type:
                            # Component found
                            user_component_found = True
                            break
                    
                    if user_component_found is False:
                        # Look for default data and load it if found
                        default_data_path = self.get_component_default_data_file(component_def.type)
                        if cg.file_exists(default_data_path):
                            # Load component from default data file
                            log_info(("Loading default data for mandatory component '%s' from " \
                                      + "file '%s'.") % (component_def.type,default_data_path))
                            component_type = component_def.type 
                            component_name = "Default: " + str(component_def.type)
                            component_desc = "Default: " + str(component_def.type)
                            component_input_type = IN_TYPE_ODS
                            component_input = default_data_path
                            component_params = {}
                            project_component = self.Component(component_type, component_input, \
                                                               component_input_type, component_params, \
                                                               component_name, component_desc)
                            
                            self.components.append(project_component)
                            log_info("Default data loaded for component '%s'." % component_name)
            log_info('Loading project components completed.')
            
            log_info("============== Resolving project paths. ==============")   
            self.resolve_paths()
            log_info("Paths expansion completed.") 
    
    #===============================================================================================    
    def load_component_data(self, component):
        """ Loads the data related to the given component and returns and object. """
        # Get module corresponding to the component type
        if component.type in self.components_definitions:
            # Get processing calvos module
            module_name = component.type
            # Add prefix for calvos package
            module_name = "calvos." + module_name
            # Call load_input function
            module = importlib.import_module(module_name)
            return_object = module.load_input(component.input_file_path, \
                                              component.input_type, component.params, self)
            component.component_object = return_object
        else:
            log_warn("Invalid component type: '%s'" % component.type)
            

    #===============================================================================================
    def generate_component_code(self, component, params = {}):
        """ Generate C-code for the given component. """
        # Check if component data has been loaded in
        if component.component_object is not None:
            # Call the code generation function for the object.
            if component.type in self.components_definitions:
                module_name = component.type
                # Add prefix for calvos package
                module_name = "calvos." + module_name
                # Call generate function
                module = importlib.import_module(module_name)
                module.generate(component.component_object, \
                         self.paths["project_output"], \
                         self.paths["project_working_dir"], \
                         self.paths["calvos_path"], \
                         params)
            else:
                log_warn("Invalid component type: '%s'" % component.type)
        else:
            log_warn("Component data has not been loaded in.") 
    
    #===============================================================================================
    def process_project(self):
        """ Processes all defined components for this project. 
        load_project function shall be called before this one. """
        log_info("============== Processing project components. ==============")
        for component in self.components:
            try:
                log_info('-------------- Loading data for component "%s" of type "%s".' \
                         % (component.name, component.type))
                # Loads user input and creates a component specific object. Such object will be
                # set in component.component_object
                self.load_component_data(component)
            except Exception as e:
                log_error('Failed to load user input for component "%s". Reason: %s' \
                          % (component.name, e) )

        log_info("============== Generating components code. ==============")
        
        
        test_str = ["rew ert wer\"$v:calvos_path$\" wert ewr", \
"wert er\"$d:test_dict,hola1$\"  erte rt \"$d:test_dict,hola1$\" tre w", \
" wert \"$func:function_name$\" tttt", \
"trey rt\"$d:test_dict,hola2$\" ttt", \
" rtyert \"$l:test_list,3$\"ttttt", \
"rrty \"$d:test_dict,hola1$\"asda\"$v:calvos_path$\"sdasd\"$l:test_list,4$\"asaaa asd  sa a dsa da\"$d:test_dict,hola4$\"", \
"hello \"$project_working$\""]
        
#        for string in test_str:
#            self.expand_all_tokens(string)
  
        # Create a pickle of all project object to be available for source generators.
        pickle_full_file_name = self.get_simple_param_val(self.module, "project_path_working") / \
            self.get_simple_param_val(self.module, "project_pickle")
        try:
            with open(pickle_full_file_name, 'wb') as f:
                pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)
        except Exception as e:
                print('Failed to create pickle file %s. Reason: %s' \
                      % (pickle_full_file_name, e))
        
        
        for component in self.components:
            try:
                log_info('-------------- Generating code for component "%s" of type "%s".' \
                         % (component.name, component.type))
                self.generate_component_code(component, component.params)
                
            except Exception as e:
                log_error('Failed to generate code for component "%s". Reason: %s' \
                          % (component.name, e) )
    
    #===============================================================================================
    def validate_given_simple_params(self, component, simple_params_list):
        """ .
        
        simple_params_list shall be a list of SimpleParam objects
        """
        all_valid = True
        valid = True
        if len(simple_params_list) > 0:
            for param in simple_params_list:
                param_id = param.param_id
                validator = param.validator
                if validator is not None:
                    log_debug("Validating parameter '%s' against its validator '%s:%s'..." \
                              % (param_id, validator[0], validator[1]))
                    valid = self.validate_simple_param(component, param_id, \
                                                       validator[0], validator[1])
                    if valid is True:
                        log_debug("Valid data.")
                    else:
                        all_valid = False
        else:
            log_debug("List of parameters to validate is empty.")
        
        return all_valid
    
    #===============================================================================================
    def validate_simple_params_defaults(self):
        """. """
        # Validate local parameters
        valid = self.validate_given_simple_params(self.module, self.simple_params.values())
        if valid is True:
            # Check each component definition's parameters
            for component in self.components_definitions.values():
                valid = self.validate_given_simple_params(component.type, \
                                                         component.simple_params.values())
                if valid is False:
                    break
        
        return valid
    
    #===============================================================================================
    def validate_comp_params(self):
        """ Validates defined component's data. """
        
        valid = True
        for component in self.components.values():
            valid = self.validate_given_simple_params(component.type, \
                                                     component.simple_params.values())
            if valid is False:
                break
        
        return valid
        
                          
    #===============================================================================================
    def simple_param_exists(self, component, param_id):
        """ Check if a component parameter is defined. """
        
        return_value = False
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                return_value = True
                
        return return_value
    
    #===============================================================================================
    def simple_param_is_read_only(self, component, param_id):
        """ Check if a component parameter is defined. """
        
        return_value = None
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                return_value = \
                    self.components_definitions[component].simple_params[param_id].read_only
                
        return return_value
    
    #===============================================================================================
    def validate_simple_param(self, component, param_id, validator_comp, validator_id):
        return_value = None
        if component in self.components_definitions \
        and param_id in self.components_definitions[component].simple_params \
        and validator_comp in self.components_definitions \
        and validator_id in self.components_definitions[component].simple_params:
            # See if param_id is within validator allowed values. Validator shall be of 'list'
            # type.
            if self.get_simple_param_type(validator_comp, validator_id) == 'list':
                param_val = \
                    self.components_definitions[component].simple_params[param_id].param_value
                validator_list = \
                 self.components_definitions[validator_comp].simple_params[validator_id].param_value
                if param_val in validator_list:
                    return_value = True
                else:
                    log_warn("Invalid data. Allowed values list: '%s'." % validator_list)
                    return_value = False
            else:
                log_warn("Validator param '%s' of component '%s' is not of 'list' type." \
                         % (validator_id, validator_comp))
            
        else:
            log_warn("Either param to be compared '%s:%s' or validator '%s:%s' not Found." \
                     % (component, param_id, validator_comp, validator_id))
        
        return return_value
    
    #===============================================================================================
    def get_simple_param_default(self, component, param_id):
        """ Get component parameter default value. """

        # Value of parameters in project's component_definitions is always the default           
        return self.get_simple_param_val(component, param_id)
    
    #===============================================================================================
    def get_simple_param_type(self, component, param_id):
        """ Get component parameter type. """
        
        return_value = None
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                return_value = \
                    self.components_definitions[component].simple_params[param_id].param_type
                
        return return_value
    
    #===============================================================================================
    def get_simple_param_validator(self, component, param_id):
        """ Get component parameter type. """
        return_value = None
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                return_value = \
                    self.components_definitions[component].simple_params[param_id].validator
                
        return return_value
    
    #===============================================================================================
    def get_simple_param_val(self, component, param_id, default = None):
        """ Gets a component parameter.
        Returns the provided default value if parameter is not found. """
        return_value = default
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                return_value = \
                    self.components_definitions[component].simple_params[param_id].param_value
            else:
                log_warn("Parameter '%s' not found in component type '%s'." \
                         % (param_id, component))
        else:
            log_warn("Component '%s' is not defined. Parameter '%s' can't be retrieved." \
                     % (component, param_id))
        
        return return_value

    #===============================================================================================
    def update_simple_param(self, component, param_id, write_value):
        """ Updates a component parameter if existing. """
        if component in self.components_definitions:
            if param_id in self.components_definitions[component].simple_params:
                self.components_definitions[component].simple_params[param_id].param_value \
                    = write_value
            else:
                log_warn("Parameter '%s' not found in component type '%'." % (component, component))
        else:
            log_warn("Component '%s' is not defined. Parameter '%s' can't be updated." \
                     % (component, param_id))
                
    
    #===============================================================================================
    TOKEN_ERROR = 0
    TOKENS_NOT_FOUND = 1
    TOKENS_REPLACED = 2
    def expand_direct_tokens(self, input_str, module = None):
        """ Retrieves local data based on the input retrieve string.
        
        retrieve string can have combination of the following elements:
          - $v:variable_name$
          - $f:function_name$
          - $d:dict_name,key$
          - $l:list_name,idx$
          - $p:module,param_id$ --> get global param_id
          - $p:module,param_id$ --> get global param_id
          - $k:string_name$ --> get named string from self.simple_params["<module>_named_strings"]
          - $alias$ --> get a new retrieve string from self.simple_params["<module>_alias_data"]
        """
        # TODO: make function also be able to provide arguments, for example with **kwargs  
        if module is None:
            module = self.module
            
        token_error = False
        current_expansion = None
        
        # Resolve direct tokens
        # ---------------------
        # Search for direct tokens $token_type:token_value$
        regex = re.compile(r"(\$(v|f|d|l|p):(.+?)\$)")
        regex_matches = re.findall(regex,str(input_str))
        
        expanded_str = input_str
        
        for match in regex_matches:
            # Element TOKEN_TYPE corresponds to the token_type, element TOKEN_VALUE to token_value
            token_full = match[0]
            token_type = match[1]
            token_value = match[2]
            current_expansion = None
             
            if token_type == "v":
                log_debug("Processing variable token '$%s:%s$'..." % (token_type, token_value))
                if hasattr(self, token_value):
                    current_expansion = getattr(self, token_value)
                    log_debug("Token expanded to: '%s'" % current_expansion)
                else:
                    log_warn("Local variable '%s' not found in object" % (token_value))
            elif token_type == "f":
                # TODO: implement this
                pass
            elif token_type == "d":
                log_debug("Processing dictionary token '$%s:%s$'..." % (token_type, token_value))
                token_val_args = token_value.split(",")
                if len(token_val_args) == 2:
                    # Element zero of splitted string is the variable name, element one is the key
                    if hasattr(self, token_val_args[0]):
                        key_str = str(token_val_args[1])
                        dict_var = getattr(self, token_val_args[0])
                        if key_str in dict_var:
                            current_expansion = dict_var[key_str]
                            log_debug("Token expanded to: '%s'" % current_expansion)
                        else:
                            log_warn("In token '$%s:%s$', key '%s' doesn't exist." \
                                     % (token_type, token_value, str(key_str)))
                    else:
                        log_warn("Local variable '%s' not found. Returned 'None'." % (token_value))
                else:
                    log_warn(("In token '$%s:%s$' separator ',' is missing or has more than one." \
                             + " Returned 'None'.") % (token_type, token_value))
            elif token_type == "l":
                log_debug("Processing list token '$%s:%s$'..." % (token_type, token_value))
                token_val_args = token_value.split(",")
                if len(token_val_args) == 2:
                    # Element zero of splitted string is the variable name, element one is the index
                    if hasattr(self, token_val_args[0]):
                        index = int(token_val_args[1])
                        list_var = getattr(self, token_val_args[0])
                        if index < len(list_var):
                            current_expansion = list_var[index]
                            log_debug("Token expanded to: '%s'" % current_expansion)
                        else:
                            log_warn("In token '$%s:%s$', index '%s' is out of range." \
                                     % (token_type, token_value, str(index)))
                    else:
                        log_warn("Local variable '%s' not found. Returned 'None'." % (token_value))
                else:
                    log_warn(("In token '$%s:%s$' separator ',' is missing or has more than one." \
                             + " Returned 'None'.") % (token_type, token_value))
            elif token_type == "p":
                log_debug("Processing parameter token '$%s:%s$'..." % (token_type, token_value))
                token_val_args = token_value.split(",")
                if len(token_val_args) == 1:
                    # Element zero of splitted string is the local param_id.
                    current_expansion = self.get_simple_param_val(self.module, token_val_args[0])
                    log_debug("Token expanded to: '%s'" % current_expansion)
                elif len(token_val_args) == 2:
                    # Element zero of splitted string is the parameter module name, element one 
                    # is the param_id.
                    current_expansion = \
                        self.get_simple_param_val(token_val_args[0], token_val_args[1])
                    log_debug("Token expanded to: '%s'" % current_expansion)
                else:
                    log_warn(("In token '$%s:%s$' separator ',' appears more than once." \
                             + " Returned 'None'.") % (token_type, token_value))
            else:
                log_warn("In token '$%s:%s$', unknown token type '%s." \
                         % (token_type, token_value, token_type))
            
            # Substitute value in input string
            if current_expansion is None:
                token_error = True
                
            # If expansion has a single back slash \ need to use a temporal set of characters
            # like '#####' since it causes problems with regex 'sub' function.
            escaping_chars = '#####'
            escape_diag_regex = re.compile(r"\\(?!\\)")
            escaped_str = escape_diag_regex.sub(escaping_chars,str(current_expansion))
            # Perform replacements
            expanded_str = regex.sub(str(escaped_str),str(expanded_str), 1)
            # Un-scape back slashes if any was escaped.
            expanded_str = expanded_str.replace(escaping_chars, "\\")
        
        output_str = expanded_str
        
        if token_error == True:
            return_stat = self.TOKEN_ERROR
        elif current_expansion is None:
            return_stat = self.TOKENS_NOT_FOUND
        else:
            return_stat = self.TOKENS_REPLACED
            
        return {"status" : return_stat, "out_str" : output_str}
        
    #===============================================================================================
    def expand_alias_tokens(self, input_str, module = None):
        """ Expands alias tokens.
        
        input_str can have combination of the following elements:
          - $alias$ --> get a new retrieve string from self.simple_params["<module>_alias_data"]
        """
        
        if module is None:
            module = self.module
            
        component_name = module.split(".")
        component_name = component_name[-1]
        
        token_error = False
        retrieve_string = None
        
        expanded_str = input_str
        
        # Resolve aliased tokens
        # ----------------------
        # Search for aliased tokens $alias$. Alias name shall comply with C-identifier syntax.
        regex = re.compile(r"(\$([a-zA-Z_][a-zA-Z0-9_]*)\$)")
        regex_matches = re.findall(regex,str(input_str))
        
        for match in regex_matches:
            # new retrieve string will be taken from get_simple_param_val(<module>, alias_data)
            token_value = match[1]
            retrieve_string = None
            
            log_debug("Processing aliased token '$%s$'..." % token_value)
            
            # Get retrieve string
            if cg.is_valid_identifier(token_value):
                alias_dict_name = str(component_name)+"_alias_dict"
                if self.simple_param_exists(module, alias_dict_name):
                    alias_dict = self.get_simple_param_val(module, alias_dict_name)
                    if token_value in alias_dict:
                        retrieve_string = str(alias_dict[token_value])
                        log_debug("Token expanded to: '%s'" % retrieve_string)
                    else:
                        log_warn("Alias name '%s' not defined. Returned 'None'. " % token_value)
                else:
                    log_warn("Alias dictionary not found for component '%s'." % module)
            else:
                log_warn("Alias name '%s' shall comply with C-identifier syntax." % token_value)
            
            if retrieve_string is None:
                token_error = True

            # Substitute value in input string
            expanded_str = regex.sub(str(retrieve_string),expanded_str, 1)

        output_str = expanded_str
    
        if token_error == True:
            return_stat = self.TOKEN_ERROR
        elif retrieve_string is None:
            return_stat = self.TOKENS_NOT_FOUND
        else:
            return_stat = self.TOKENS_REPLACED
        
        return {"status" : return_stat, "out_str" : output_str}
    
    #===============================================================================================
    def expand_all_tokens(self, input_str, module = None):
        """ Expand all replacing tokens in the given string. """
        
        if module is None:
            module = self.module
        
        tokens_found = False
        token_error = False
        
        # First expand direct tokens then expand aliased tokens and finally expand again possible
        # direct tokens generated while expanding aliased tokens. The order is important since
        # one first replacement of direct tokens needs to occur before the aliased tokens to
        # avoid confusing 'non-token' text in between tokens as if they were aliased tokens.
        log_debug("Expanding string: '%s'..." % input_str)
        return_value = self.expand_direct_tokens(input_str, module)
        if return_value["status"] == self.TOKENS_REPLACED:
            tokens_found = True
            log_debug("String with direct tokens replaced: '%s'" % return_value["out_str"])
        elif return_value["status"] == self.TOKEN_ERROR:
            token_error = True
        
        # Keep expanding all aliased/direct tokens until no one is found
        while_guard = grl.LOOP_GUARD_MED
        last_loop = False
        while while_guard > 0:
            while_guard -= 1
            if while_guard == 0:
                log_error(("Error expanding tokens. Search took %s loops without success " \
                          + "and was canceled.") % str(grl.LOOP_GUARD_MED))
                
            alias_tokens_this_loop = False
            direct_tokens_this_loop = False
                
            return_value = self.expand_alias_tokens(return_value["out_str"], module)
            
            if return_value["status"] == self.TOKEN_ERROR:
                token_error = True
                break
            elif return_value["status"] == self.TOKENS_REPLACED:
                tokens_found = True
                alias_tokens_this_loop = True
                log_debug("Expanded string: '%s'" % return_value["out_str"])
            
            # Process direct tokens resulting from expansion of aliased ones
            return_value = self.expand_direct_tokens(return_value["out_str"], module)
            if return_value["status"] == self.TOKEN_ERROR:
                token_error = True
                break
            elif return_value["status"] == self.TOKENS_REPLACED:
                tokens_found = True
                direct_tokens_this_loop = True
                
            if alias_tokens_this_loop is True or direct_tokens_this_loop is True:
                last_loop = False
            elif alias_tokens_this_loop is False and direct_tokens_this_loop is False:
                if last_loop is True:
                    break
                else:
                    last_loop = True
            
            
        # Determine returning values
        if token_error == True:
            return_value["status"] = self.TOKEN_ERROR
            log_warn("Final expanded string with errors: '%s'" % return_value["out_str"])
        elif tokens_found == True:
            return_value["status"] = self.TOKENS_REPLACED
            log_debug("Final expanded string: '%s'" % return_value["out_str"]) 
        else:
            return_value["status"] = self.TOKENS_NOT_FOUND
            log_debug("No replacements performed.")
            
        return return_value
            
    #===============================================================================================
    def resolve_paths(self):
        """ Expands all parameters of type "path". """ 
        for param in self.simple_params.values():
            if param.param_type == "path":
                current_path = self.expand_all_tokens(param.param_value)["out_str"]
                current_path = cg.string_to_path(current_path)
                param.param_value = current_path                
                log_debug("Resolved path '%s': '%s'" % (param.param_id, param.param_value))
                
        for comp_def in self.components_definitions.values():
            for param in comp_def.simple_params.values():
                if param.param_type == "path":
                    current_path = self.expand_all_tokens(param.param_value)["out_str"]
                    current_path = cg.string_to_path(current_path)
                    param.param_value = current_path  
                    log_debug("Resolved component '%s' path '%s': '%s'" \
                              % (comp_def.type, param.param_id, param.param_value))
    
    #===============================================================================================
    def get_work_file_path(self, comp_id, param_id):
        """ Returns a Path to the working directory file from a parameter. 
        File name is extracted from a the parameter comp_id:param_id. """
        
        return_value = None
        work_dir = self.get_simple_param_val(self.module,"project_path_working")
        work_file = self.get_simple_param_val(comp_id,param_id)
        if work_dir is not None and work_file is not None:
            return_value = work_dir / work_file
        
        return return_value
    
    #===============================================================================================
    def get_component_root_path(self, component_id):
        """ Returns the path with cog files for the given component. 
        
        Parameters
        ----------
            component_id, str
                String constituing the component name, for example: "utils.time" or
                "comgen.CAN", etc.
        """     
        path_lst = component_id.split(".")
        module_name = path_lst[-1]
        root_path_val = cg.string_to_path(self.calvos_path)
        for i, element in enumerate(path_lst):
            if i < len(path_lst) - 1:
                root_path_val = root_path_val / element
                
        return root_path_val
    
    #===============================================================================================
    def get_component_gen_path(self, component_id):
        """ Returns the path with cog files for the given component. 
        
        Parameters
        ----------
            component_id, str
                String constituing the component name, for example: "utils.time" or
                "comgen.CAN", etc.
        """     
        gen_path_lst = component_id.split(".")
        module_name = gen_path_lst[-1]
        gen_path_val = cg.string_to_path(self.calvos_path)
        for i, element in enumerate(gen_path_lst):
            if i < len(gen_path_lst) - 1:
                gen_path_val = gen_path_val / element
        gen_path_val = gen_path_val / "gen" / module_name
        
        return gen_path_val
    
    #===============================================================================================
    def get_component_default_data_file(self, component_id):
        """ Returns the expected file name with default data for mandatory components.
        
        Parameters
        ----------
            component_id, str
                String constituting the component name, for example: "utils.time" or
                "comgen.CAN", etc.
        """     
        data_path_lst = component_id.split(".")
        module_name = data_path_lst[-1]
        data_path_val = cg.string_to_path(self.calvos_path)
        file_name = "default_data_"
        for i, element in enumerate(data_path_lst):
            if i < len(data_path_lst) - 1:
                data_path_val = data_path_val / element
                file_name += str(element) + "-"   
        data_path_val = data_path_val / "usr_in" / (file_name + module_name + ".ods")
        
        return data_path_val
    
    #===============================================================================================
    def get_list_of_templates(self):
        """ Returns a list of paths with file names of user in templates """
        component_lst = self.find_components(self.calvos_path)
        comp_paths = []
        template_files = []
        
        # Gather list of component's root paths
        for component_id in component_lst.keys():
            comp_path = self.get_component_root_path(component_id)
            comp_path /= "usr_in"
            if comp_path not in comp_paths:
                comp_paths.append(comp_path)
        
        # Gather all templates from path list created above
        for path in comp_paths: 
            # Find all files starting with "template -"
            for file_path in path.glob('template -*.*'):
                if file_path.is_file() is True:
                    if file_path not in template_files:
                        template_files.append(file_path)
        
        return template_files
    
    #===============================================================================================
    def gen_demo_project(self, out_path):
        """ Generates a Demo Project and copies it to the provided output path. """

        if cg.folder_exists(out_path):
            # Gather demo project template and clean it up
            # --------------------------------------------
            # Get XML
            nsmap = {"clv":"calvos"}
            
            ET.register_namespace("clv","calvos")
            
            XML_TEMPLATE_NAME = "calvos_project.xml"
            project_xml = self.get_component_root_path(self.module) / "usr_in" / XML_TEMPLATE_NAME
            XML_tree = ET.parse(project_xml)
            XML_root = XML_tree.getroot()
                 
            xml_components = XML_root.find("Components")
            # Remove any defined component -dummy ones-
            for components in XML_root.findall("./Components"):
                for component in components:
                    components.remove(component)
            
            # Load parameters of found components
            # -----------------------------------
            log_info('---------------- Loading project parameters and component definitions...')
            components_lst = self.find_components(self.calvos_path)
            for comp_name, comp_xml in components_lst.items():
                # Create component definition object
                self.components_definitions.update({comp_name : self.CompDefinition(comp_name)})
                # Parse XML with component definitions
                defs_XML_tree = ET.parse(comp_xml)
                defs_XML_root = defs_XML_tree.getroot()
                self.load_component_definitions(defs_XML_root, self.components_definitions[comp_name])
            
            log_info('Loaded project parameters and component definitions.')
            
            # Update XML project with found components
            # ----------------------------------------
            log_info('---------------- Creating demo project XML...')
            for component_id in components_lst.keys():
                # Check if a template user input is defined. Only components with a defined user
                # input will become an XML project component
                if len(self.components_definitions[component_id].inputs) > 0:
                    root_path = self.get_component_root_path(component_id)
                    log_debug("Creating Component XML element for component '%s'..." % component_id)
                    # Create component in XML
                    xml_component = ET.SubElement(xml_components, "Component")
                    xml_component.set("type", component_id)
                    xml_subelement = ET.SubElement(xml_component, "Name")
                    xml_subelement.text = self.components_definitions[component_id].title
                    xml_subelement = ET.SubElement(xml_component, "Desc")
                    xml_subelement.text = self.components_definitions[component_id].desc
                    for input_file, file_type \
                    in self.components_definitions[component_id].inputs.items():
                        # Add input to XML    
                        in_file = root_path / "usr_in" / str(input_file)
                        usr_in_file = "\"usr_in/"+str(input_file)+"\""
                        # Add input file to project XML
                        xml_subelement = ET.SubElement(xml_component, "Input")
                        xml_subelement.set("type", file_type)
                        xml_subelement.text = usr_in_file
                        # Export file
                        log_debug("Exporting user input template file '%s'..." % str(input_file))
                        if cg.folder_exists(out_path / "usr_in") is False:
                            cg.create_folder(out_path / "usr_in")
                        dest_file = out_path / "usr_in" / str(input_file)
                        shutil.copy(in_file, dest_file)
                        
                        # Export any found markdown document
                        if cg.folder_exists(out_path / "doc") is False:
                            cg.create_folder(out_path / "doc")
                        doc_path = root_path / "doc"
                        for doc_file in doc_path.glob('*.md'):
                            if doc_file.is_file() is True:
                                print("File: ", doc_file)
                                dest_file = out_path / "doc" / doc_file.name
                                # copy to demo project doc folder
                                shutil.copy(doc_file, dest_file)
                        
            # Add comment to do not modify metadata
            comment = ET.Comment(' Do NOT modify these metadata information. ')
            xml_metadata = XML_root.find("Metadata")
            if xml_metadata is not None:
                xml_metadata.insert(0,comment)
            
            
            log_info('Demo project XML created.')
            log_info('---------------- Exporting Demo project XML...')
            # Export xml project file
            out_path = cg.string_to_path(out_path)
            tree = ET.ElementTree(XML_root)
            self.indent_xml(XML_root)  
            tree.write((out_path / XML_TEMPLATE_NAME), encoding="UTF-8")
            log_info('Demo project XML exported.')
        else:
            log_error("Provided folder '%s' doesn't exist." % out_path)
                  
    #===============================================================================================
    def indent_xml(self, elem, level=0):
        """ Indents an XML root element.
        Taken from stack overvlow:
        https://stackoverflow.com/questions/3095434/...
        inserting-newlines-in-xml-file-generated-via-xml-etree-elementtree-in-python
        """
        
        i = "\n" + level*"  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent_xml(elem, level+1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i
    
    #===============================================================================================        
    class CompDefinition:
        """ Class for modeling a calvos project component definition. """
        def __init__(self, type_name, **kwargs):
            self.type = type_name
            
            self.title = kwargs.get('title', None)
            self.desc = kwargs.get('desc', None)
            self.instances = kwargs.get('instances', None)
            self.params = kwargs.get('params', None)
            self.simple_params = {} # {param_id : SimpleParam object}
            self.inputs = {}    # {file_name : file_type ("ods", "xml")}
            self.default_inputs = {}    # {file_name : file_type ("ods", "xml")}
    
    #===============================================================================================
    class Component:
        """ Models a calvos project Component. """
        
        component_idx = 0
        
        #===========================================================================================
        def __init__(self, type_name, \
                     input_file = None, input_type = IN_TYPE_XML, params = {}, \
                     name = None, desc = None):
            
            self.comp_idx = self.get_and_increase_comp_idx()
            
            self.name = str(name)
            self.desc = str(desc)
            
            self.type = type_name
            
            self.input_type = input_type
            self.input_file = input_file
            self.input_file_path = None
            
            if input_file is not None:
                self.set_input(input_file, input_type)
                
            self.params = params
            
            self.component_object = None
            
            self.simple_params = {}
        
        #===========================================================================================
        @classmethod
        def get_and_increase_comp_idx(cls):
            """ Returns current class component index and then increase it by 1. """
            
            return_value = cls.component_idx
            cls.component_idx += 1
            return return_value
        
        #===========================================================================================    
        def load_param_defaults(self):
            """ Loads the default parameter values for this component type. """
            pass
        
        #===========================================================================================    
        def add_param(self, param_name, value):
            """ Adds/updates the given parameter to this component. """
            self.params.update({param_name : value})
        
        #===========================================================================================
        def remove_param(self, param_name):
            """ Removes a parameter from this component. """
            result = self.params.pop(param_name, None)
            if result is None:
                log_warn("Parameter \"%s\" to be removed didn't exist." % param_name)
        
        #===========================================================================================
        def set_input(self, input_file, input_type = IN_TYPE_XML):
            """ Sets the input file and input type for this component. """
            self.input_file_path = cg.string_to_path(str(input_file))
            
            if input_type == IN_TYPE_XML or input_type == IN_TYPE_ODS:
                self.input_type = input_type
            else:
                self.input_type = None
                log_error("Wrong input type parameter for project component: %s" % input_type)
        
        #===========================================================================================
        def set_object(self, component_object):
            self.component_object = component_object 
                        


# -*- coding: utf-8 -*-
""" CalvOS General Module.

Module for miscelaneous general features of the calvos engine.

@author: Carlos Calvillo
@copyright:  2020 Carlos Calvillo. All rights reserved.
@license:    GPL v3
"""
__version__ = '0.0.1'
__date__ = '2020-12-20'
__updated__ = '2020-12-20'

import json
import re

import calvos.common.codegen as cg
import calvos.common.logsys as lg


# --------------------------------------------------------------------------------------------------
# Definitions for the logging system
# --------------------------------------------------------------------------------------------------
LOGGER_LABEL = "grl"

log = lg.log_system
log.add_logger(LOGGER_LABEL)

def log_debug(message):
    """ Wrapper for logging a debug message. """
    
    log.debug(LOGGER_LABEL, message)
    
def log_info(message):
    """ Wrapper for logging an info message. """
    
    log.info(LOGGER_LABEL, message)
    
def log_warn(message):
    """ Wrapper for logging a warning message. """
    
    log.warning(LOGGER_LABEL, message)
    
def log_error(message):
    """ Wrapper for logging an error message. """
    
    log.error(LOGGER_LABEL, message)
    
def log_critical(message):
    """ Wrapper for logging a critical error message. """
    
    log.critical(LOGGER_LABEL, message)
# --------------------------------------------------------------------------------------------------

LOOP_GUARD_MED = 500

TYPE_SCALAR = 0
TYPE_ENUM = 1
TYPE_STRING = 2

def process_simple_param(param_type, param_value):
    """ Returns a variable corresponding to the XML parameter string according to its type. """
    
    error_msg = ""
    if param_type == "int":
        try:
            if cg.is_hex_string(param_value):
                param_value = int(str(param_value),0)
            else:
                param_value = int(json.loads(param_value))
        except Exception as e:
            param_value = None
            error_msg = "Expected int value for given parameter. Assumed 'None'.\n" \
                + "Error: " + str(e)
    elif param_type == "float":
        try:
            param_value = float(json.loads(param_value))
        except Exception as e:
            param_value = None
            error_msg = "Expected float value for given parameter. Assumed 'None'.\n" \
                + "JSON error: " + str(e)
    elif param_type == "str" or param_type == "path":
        param_value = str(param_value)
    elif param_type == "list":
        try:
            param_value = list(json.loads(param_value))
        except Exception as e:
            param_value = None
            error_msg = "Expected list type value for given parameter. Assumed 'None'.\n" \
                + "JSON error: " + str(e)
    elif param_type == "dict":
        try:
            param_value = dict(json.loads(param_value))
        except Exception as e:
            param_value = None
            error_msg = "Expected dictionary value for given parameter. Assumed 'None'.\n" \
                + "JSON error: " + str(e)
    elif param_type == "boolean":
        if param_value == "True" or param_value == "true" \
        or param_value == "TRUE" or param_value == True:
            param_value = True
        elif param_value == "False" or param_value == "false" \
        or param_value == "FALSE"  or param_value == False:
            param_value = False
        else:
            param_value = None
            error_msg = "Expected boolean value for given parameter. Assumed 'None'."
    elif param_type == "cid":
        param_value = str(param_value)
        if cg.is_valid_identifier(param_value) is False:
            param_value = None
            error_msg = "Expected C-identifier value for given parameter. Assumed 'None'."
    elif param_type == "enum":
        param_value = cg.parse_special_string(param_value)
        if len(param_value) == 0:
            param_value = None
            error_msg ="Expected special string value for given parameter. Assumed 'None'."
    else:
        param_value = None
        error_msg = "Unknown parameter type '%s' for given parameter. Assumed 'None'." % param_type
    
    # Throw warning if any
    if error_msg != "":
        log_warn(error_msg)
        
    return param_value

class SimpleParam():
    def __init__(self, param_id, param_type, param_value = None, read_only = True, \
                 validator = None):
        self.param_id = param_id
        self.param_type = param_type
        self.param_value = param_value
        self.read_only = read_only
        self.validator = validator

class Parameter():
    
    def __init__(self, id, value = None, default = None, \
                 name = "", description = ""):
        self.id = id
        self.value = value
        self.default = default
        self.name = name
        self.desc = description
        
        self.category = None
        
        self.enumeration = {}
        self.type = TYPE_SCALAR
        self.multiple = False
        
        
        self.min = None
        self.max = None
        self.offset = None
        self.resolution = None
        
    def is_valid_value(self, value = None):
        """ validate if the given value is within the defined enum. """
        return_value = False
        
        if value is None:
            # Validate current parameter value
            if self.value in self.enumeration:
                return_value = True
        else:
            # Validate passed value
            if value in self.enumeration:
                return_value = True 
        
        return return_value

class ParamCategory():
    
    def __init__(self, id, name = "", description = ""):
        self.id = id
        self.name = name
        self.desc = description


#!/usr/local/bin/python2.7
# encoding: utf-8
'''
CalvOS - Open Source SW Utilitites for Embedded Systems.

Calvos provides a set of SW utilitites for the development
of embedded systems in C-language.

@author:     Carlos Calvillo

@copyright:  2020 Carlos Calvillo. All rights reserved.

@license:    GPL v3

@contact:    calcore.io@gmail.com
@deffield    updated: Updated
'''           
import sys
import os
import pathlib as pl
import shutil
import logging
import traceback

from argparse import ArgumentParser
from argparse import RawDescriptionHelpFormatter

__all__ = []
__date__ = '2020-12-18'
__updated__ = '2020-12-18'

DEBUG = 0
TESTRUN = 0
PROFILE = 0

class CLIError(Exception):
    '''Generic exception to raise and log different fatal errors.'''
    def __init__(self, msg):
        super(CLIError).__init__(type(self))
        self.msg = "E: %s" % msg
    def __str__(self):
        return self.msg
    def __unicode__(self):
        return self.msg


def string_to_path(path_string):
    ''' Returns a "path" object for the given string. '''
    # pathlib requires regular diagonal, not reverse diagonal
    path_string = path_string.replace("\\", "/")
    
    return pl.Path(path_string)
    
    
def folder_exists(path):
    ''' Returns "True" if the given path is an existing folder. '''
    return_value = False
    
    if type(path) is str:
        path = string_to_path(path)
     
    if path.exists() and path.is_dir():
        return_value = True
    
    return return_value


def file_exists(path):
    ''' Returns "True" if the given path is an existing file. '''
    return_value = False
    
    if type(path) is str:
        path = string_to_path(path)
     
    if path.exists() and path.is_file():
        return_value = True
    
    return return_value  
    

def main(argv=None): # IGNORE:C0111
    '''Command line options.'''
    # Import version from __version__.py file
    # Set calvos path which will be used for accessing package resources
    calvos_path = pl.Path(__file__).parent.absolute()
    version_path = calvos_path / "__version__.py"
        
    about = {}
    with open(version_path) as f:
        exec(f.read(), about)
    __version__ = about["__version__"]

    if argv is None:
        argv = sys.argv
    else:
        sys.argv.extend(argv)

    program_name = os.path.basename(sys.argv[0])
    program_version = "v%s" % __version__
    program_build_date = str(__updated__)
    program_version_message = 'calvos %s' % (program_version)
    program_shortdesc = __import__('__main__').__doc__.split("\n")[1]
    program_license = '''%s
Copyright (C) 2020  Carlos Calvillo
  
  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.  

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <https://www.gnu.org/licenses/>.

USAGE
''' % program_shortdesc

    try:
        # Setup argument parser
        parser = ArgumentParser(description=program_license, \
                                formatter_class=RawDescriptionHelpFormatter)
        parser.add_argument("-d","--demo", dest="demo", required=False, \
            help=("Will provide an example calvos project with user input templates in " \
                +"the given DEMO path. No project will be processed if this argument is provided."))
        parser.add_argument('-V', '--version', action='version', version=program_version_message)
        parser.add_argument('-v', '--ver', action='version', version=program_version_message)
        parser.add_argument("-p","--project", dest="project", \
            required=(('--demo' not in sys.argv) and ('-d' not in sys.argv)), \
            help=("Required (if -d was not provided). Full path with file name of the calvos" \
                  + "project to be processed."))
        parser.add_argument("-l","--log", dest="log_level", required=False, \
            help=("Optional. LOG_LEVEL: 0 - Debug, 1 - Info, "\
                  + "2 - Warning, 3 - Error. Default is 1 - Info."))
        parser.add_argument("-e","--export", dest="export", required=False, \
            help="Optional. Generated C-code will be exported (copied) into the provided EXPORT path.")
        parser.add_argument("-b","--backup", dest="backup", required=False, \
            help=("Optional. Backups of the overwritten C-code during an export operation " \
                  + "will be placed in the provided BACKUP path. " \
                  + "This is only used if -e argument was provided."))
        

        # Process arguments
        args = parser.parse_args()
        project = args.project
        log_level = args.log_level
        
        arguments_OK = True
        
        # Check if calvos package is installed by trying to import the logging system
        try:
            import calvos.common.logsys as lg
        except:
            print("INFO: Setting up package...")
            # Add package path to sys.path and attempt the importing again
            calvos_pkg_path_str = str(calvos_path.parent)
            if calvos_pkg_path_str not in sys.path:
                sys.path.append(calvos_pkg_path_str)
                print("INFO: calvos path added to sys.path.")
            # Attempt the import
            try:
                import calvos.common.logsys as lg
            except Exception as e:
                print("ERROR: Couldn't load calvos package.")
                print("Generated exception:\n")
                print(str(e))
                return 2
        
        if log_level is not None:
            if int(log_level) == 0:
                log_level = logging.DEBUG
            elif int(log_level) == 1:
                log_level = logging.INFO
            elif int(log_level) == 2:
                log_level = logging.WARNING
            elif int(log_level) == 3:
                log_level = logging.ERROR
            else:
                arguments_OK = False
                print('Error: argument -l shall be from 0 to 3. Provided value: "%s"',log_level)
                return 2
        else:
            log_level = logging.INFO
            
        #==============================================================================
        # Generate demo project if argument -t is provided.
        #==============================================================================
        if args.demo is not None:
            # Check if path exists
            demo_path = string_to_path(args.demo)
            if folder_exists(demo_path):
                
                # Setting up logging system
                log_output_file = demo_path / "log.log"
                if file_exists(log_output_file):
                    log_output_file.unlink()
                lg.log_system = lg.Log(log_level, log_output_file)
                log = lg.log_system
                log.add_logger("main")
                
                # Create project object
                import calvos.common.project as pj
                
                log.info("main", \
                         "============== Generating Demo Project ==============")
                print("INFO: ============== Generating Demo Project ==============")
                calvos_project = pj.Project("Project Name", None, calvos_path)
                calvos_project.gen_demo_project(demo_path)
                print("INFO: Done. Demo project generated. ")
                log.info("main", "Done. Demo project generated. ")
            else:
                print("ERROR: Provided folder '%s' doesn't exist." % demo_path)
                return 2
                
            print("INFO: Don't provide argument -d if a project needs to be processed.")
            return 0    
        
        #==============================================================================
        # Start project processing if argument -t was not provided
        #==============================================================================
        if project is not None:
            project_file = string_to_path(project)
            if file_exists(project_file):
                # Extract project path
                project_path = project_file.parent.resolve()
                print("INFO: Project path: ", project_path)
            else:
                arguments_OK = False
                print("Error: file: ", project_file, " doesn't exists.")

        if arguments_OK is True:
            #==============================================================================
            # Setup logging system
            #==============================================================================
            log_output_file = project_path / "log.log"
            if file_exists(log_output_file):
                log_output_file.unlink()
             
            lg.log_system = lg.Log(log_level, log_output_file)
            log = lg.log_system
            log.add_logger("main")
            log.info("main","============== Started calvOS project processing. ==============")
            print("INFO: ============== Started calvOS project processing. ==============")
             
            #==============================================================================
            # calvOS project folders
            #==============================================================================           
            project_path_user_inputs = project_path / "usr_in"
            project_path_output = project_path / "out"
            project_path_working_dir = project_path / "out/working_dir"
            project_path_docs = project_path / "out/doc"
            project_path_templates = project_path / "out/doc/templates"
            
            import calvos.common.codegen as cg
 
            # Set global variables for code generation
            cg.calvos_path = calvos_path
            cg.calvos_project_path = project_path
            
            #==============================================================================
            # Setup needed calvOS project folders
            #==============================================================================
            log.info("main","Setting up project folders.")
            #Delete output and working directories
            cg.delete_folder_contents(project_path_output)
              
            # Recreate output folder structure
            cg.create_folder(project_path_output)
            cg.create_folder(project_path_working_dir) 
            cg.create_folder(project_path_docs) 
            cg.create_folder(project_path_templates)  

            
            #==============================================================================
            # Process project
            #==============================================================================
            import calvos.common.project as pj

            log.info("main","============== Loading project infrastructure. ==============")
            
            log.info("main",'Create instance for project in "%s"' % project_file)
            calvos_project = pj.Project("Project Name", project_file, calvos_path)
            
            calvos_project.load_project()
            
            calvos_project.process_project()
            
            #==============================================================================
            # Process export argument
            #==============================================================================
            if args.export is not None:
                log.info("main","============== Exporting generated code. ==============")
                print("INFO: Exporting generated C-code...")
                export_path = string_to_path(args.export)
                if cg.folder_exists(export_path):
                    # Get files to export
                    c_files = list(project_path_output.glob('*.c'))
                    h_files = list(project_path_output.glob('*.h'))
                    
                    # If backup location is defined then do the backup
                    if args.backup is not None:
                        MAX_BKUP_COPIES = 10
                        current_bkup_copy = 0
                        
                        log.info("main","-------------- Backing up files. --------------")
                        print("INFO: Backing up C-code to be overwritten during export...")
                        backup_path =  string_to_path(args.backup)
                        if export_path != backup_path:
                            if cg.folder_exists(backup_path) is False:
                                log.info("main", " Backup path '%s' doesn't exist, creating it..." \
                                         % backup_path)
                                cg.create_folder(backup_path)
                            # Get backup copy number to use  
                            for file in c_files:
                                while current_bkup_copy <= MAX_BKUP_COPIES:
                                    file_name = str(file.name)
                                    if current_bkup_copy > 0:
                                        file_name = str(current_bkup_copy) + "_" + file_name  
                                    dest_file = backup_path / file_name
                                    if cg.file_exists(dest_file):
                                        # Increment copy number  
                                        current_bkup_copy += 1
                                    else:
                                        # File doesn't exist, use current backup cupy number
                                        break
                            
                            # Backup files to be overwritten
                            file_counter = 0
                            for file in c_files:
                                file_name = str(file.name)
                                file_name_dest = file_name
                                if current_bkup_copy > 0:
                                    file_name_dest = str(current_bkup_copy) + "_" +  file_name_dest
                                orig_file = export_path / file_name
                                dest_file = backup_path / file_name_dest
                                if cg.file_exists(orig_file):
                                    # Copy file to backup location
                                    shutil.copy(orig_file, dest_file)
                                    file_counter += 1
                            for file in h_files:
                                file_name = str(file.name)
                                file_name_dest = file_name
                                if current_bkup_copy > 0:
                                    file_name_dest = str(current_bkup_copy) + "_" + file_name_dest 
                                orig_file = export_path / file_name
                                dest_file = backup_path / file_name_dest
                                if cg.file_exists(orig_file):
                                    # Copy file to backup location
                                    shutil.copy(orig_file, dest_file)
                                    file_counter += 1
                            
                            log.info("main", " Backup completed. '%s' file(s) backed-up." \
                                         % file_counter)
                            print("INFO: Backup completed. '%s' file(s) backed-up." % file_counter)
                        else:
                            arguments_OK = False
                            log.warning("main", "Provided Export and Backup folders are same. " \
                                + "Export operation cancelled.")
                    
                    # Export files
                    if arguments_OK is True:
                        log.info("main","-------------- Exporting files. --------------")
                        
                        file_counter = 0
                        for file in c_files:
                            file_name = file.name
                            orig_file = project_path_output / file_name
                            dest_file = export_path / file_name
                            if cg.file_exists(orig_file):
                                # Copy file to backup location
                                shutil.copy(orig_file, dest_file)
                                file_counter += 1
                        for file in h_files:
                            file_name = file.name
                            orig_file = project_path_output / file_name
                            dest_file = export_path / file_name
                            if cg.file_exists(orig_file):
                                # Copy file to backup location
                                shutil.copy(orig_file, dest_file)
                                file_counter += 1
                        
                        log.info("main", "Export completed. '%s' file(s) exported." % file_counter)
                        print("INFO: Export completed. '%s' file(s) exported." % file_counter)
                else:
                    log.warning("main", "Export folder '%s' doesn't exist. No export performed." \
                             % export_path)
                       
            log.info("main","============== Finished calvOS project processing. ==============")
            print("INFO: ============== Finished calvOS project processing. ==============")
            for counter_name, counts in log.counters.items():
                log.info("main", "\"" + counter_name + "\" messages : " + str(counts))
                print("INFO: \"" + counter_name + "\" messages : " + str(counts))
            print("INFO: See log file '%s' for details." % str(log_output_file))
              
            logging.shutdown()
   
        return 0
    except KeyboardInterrupt:
        ### handle keyboard interrupt ###
        return 0
    except Exception as e:
        if DEBUG or TESTRUN:
            raise(e)
        indent = len(program_name) * " "
        
        message = program_name + ": " + repr(e) + "\n"
        message = message + traceback.format_exc()
        sys.stderr.write(program_name + ": " + repr(e) + "\n")
        sys.stderr.write(indent + "  for help use --help")
        
        if 'log' in locals():
            log.critical("main",message)
        
        string_ms = traceback.print_exc()
        
        return 2

if __name__ == "__main__":
    if DEBUG:
        sys.argv.append("-p")
        sys.argv.append('G:\\devproj\\github\\calvos\\calvos\\project_example\\project.xml')
        sys.argv.append("-c")
        sys.argv.append('G:\\devproj\\github\\calvos\\calvos\\calvos-engine\\calvos')
    if TESTRUN:
        import doctest
        doctest.testmod()
    if PROFILE:
        import cProfile
        import pstats
        profile_filename = 'calvos.main_profile.txt'
        cProfile.run('main()', profile_filename)
        statsfile = open("profile_stats.txt", "wb")
        p = pstats.Stats(profile_filename, stream=statsfile)
        stats = p.strip_dirs().sort_stats('cumulative')
        stats.print_stats()
        statsfile.close()
        sys.exit(0)
    sys.exit(main())


'''
Created on 26 ene. 2021

@author: Carlos Calvillo
'''
import calvos.common.codegen as cg
import calvos.common.logsys as lg

# --------------------------------------------------------------------------------------------------
# Definitions for the logging system
# --------------------------------------------------------------------------------------------------
LOGGER_LABEL = "cog_time"

log = lg.log_system
log.add_logger(LOGGER_LABEL)

def log_debug(message):
    """ Wrapper for logging a debug message. """
    
    log.debug(LOGGER_LABEL, message)
    
def log_info(message):
    """ Wrapper for logging an info message. """
    
    log.info(LOGGER_LABEL, message)
    
def log_warn(message):
    """ Wrapper for logging a warning message. """
    
    log.warning(LOGGER_LABEL, message)
    
def log_error(message):
    """ Wrapper for logging an error message. """
    
    log.error(LOGGER_LABEL, message)
    
def log_critical(message):
    """ Wrapper for logging a critical error message. """
    
    log.critical(LOGGER_LABEL, message)
# --------------------------------------------------------------------------------------------------
 
def C_gen_info(worksheet, network_object):
    """ Returns a C-comment with network information. """
    # Adjust file string "worksheet" to do not exceed 80 characters and to start at 
    # a given character
    file_str = cg.str_to_multiline(worksheet, (80-22), 22)   
    # Print generation information
    curren_time = cg.get_local_time_formatted() # Gets current local time
    return_str = """/*-----------------------------------------------------------------------------
 * This file was generated on (yyyy.mm.dd::hh:mm:ss): """ + curren_time +"""
 -----------------------------------------------------------------------------*/"""
    return return_str